@@ -1,11 +1,11 @@
 ;;; dired-preview.el --- Automatically preview file at point in Dired -*- lexical-binding: t -*-
 
-;; Copyright (C) 2023  Free Software Foundation, Inc.
+;; Copyright (C) 2023-2024  Free Software Foundation, Inc.
 
 ;; Author: Protesilaos Stavrou <info@protesilaos.com>
-;; URL: https://git.sr.ht/~protesilaos/dired-preview
-;; Mailing-List: https://lists.sr.ht/~protesilaos/general-issues
-;; Version: 0.1.1
+;; Maintainer: Protesilaos Stavrou <info@protesilaos.com>
+;; URL: https://github.com/protesilaos/dired-preview
+;; Version: 0.2.0
 ;; Package-Requires: ((emacs "27.1"))
 ;; Keywords: files, convenience
 
@@ -57,6 +57,10 @@
 ;;; Code:
 
 (require 'dired)
+(require 'seq)
+(eval-when-compile
+  (require 'cl-lib)
+  (require 'subr-x))
 
 (defgroup dired-preview nil
   "Automatically preview file at point in Dired."
@@ -64,17 +68,25 @@
 
 (defcustom dired-preview-ignored-extensions-regexp
   (concat "\\."
-          "\\(mkv\\|webm\\|mp4\\|mp3\\|ogg\\|m4a"
+          "\\(mkv\\|webm\\|mp4\\|mp3\\|ogg\\|m4a\\|flac\\|wav"
           "\\|gz\\|zst\\|tar\\|xz\\|rar\\|zip"
           "\\|iso\\|epub\\|pdf\\)")
-  "Regular expression of file type extensions to not preview."
+  "Regular expression of file type extensions to not preview.
+When the value is nil, do not ignore any file: preview
+everything."
   :group 'dired-preview
-  :type 'string)
+  :type '(choice (const :tag "Do not ignore any file (preview everything)" nil)
+                 (string :tag "Ignore files matching regular expression")))
+
+(defcustom  dired-preview-image-extensions-regexp "\\.\\(png\\|jpg\\|jpeg\\|tiff\\)"
+  "List of file extensions representing image types."
+  :group 'dired-preview
+  :type '(string :tag "Image files matching regular expression"))
 
 (defcustom dired-preview-max-size (expt 2 20)
   "Files larger than this byte limit are not previewed."
   :group 'dired-preview
-  :type 'integer)
+  :type 'natnum)
 
 (defcustom dired-preview-display-action-alist-function
   #'dired-preview-display-action-alist-dwim
@@ -90,19 +102,22 @@ details."
 (defcustom dired-preview-delay 0.7
   "Time in seconds to wait before previewing."
   :group 'dired-preview
-  :type 'float)
+  :type 'number)
+
+(defcustom dired-preview-chunk-size 10240
+  "Size in bytes to read from large files."
+  :group 'dired-preview
+  :type 'natnum)
 
 (defvar dired-preview--buffers nil
   "List with buffers of previewed files.")
 
+(defvar dired-preview--large-files-alist nil
+  "Alist mapping previewed large files to buffer names.")
+
 (defun dired-preview--get-buffers ()
   "Return buffers that show previews."
-  (seq-filter
-   (lambda (buffer)
-     (when (and (bufferp buffer)
-                (buffer-live-p buffer))
-       buffer))
-   dired-preview--buffers))
+  (seq-filter #'buffer-live-p dired-preview--buffers))
 
 (defun dired-preview--window-parameter-p (window)
   "Return non-nil if WINDOW has `dired-preview-window' parameter."
@@ -132,16 +147,23 @@ until it drops below this number.")
     (catch 'stop
       (mapc
        (lambda (buffer)
-         (if (and (>= (dired-preview--get-buffer-cumulative-size)
-                      dired-preview--buffers-threshold))
-             (when (and (buffer-local-value 'delayed-mode-hooks buffer)
-                        (not (eq buffer (current-buffer))))
-               (ignore-errors (kill-buffer-if-not-modified buffer))
-               (setq buffers (delq buffer buffers)))
-           (throw 'stop t)))
+         (when (and (>= (dired-preview--get-buffer-cumulative-size) dired-preview--buffers-threshold)
+                    (not (eq buffer (current-buffer))))
+           (ignore-errors (kill-buffer-if-not-modified buffer))
+           (setq buffers (delq buffer buffers)))
+         (throw 'stop t))
        buffers))
     (setq dired-preview--buffers (delq nil (nreverse buffers)))))
 
+(defun dired-preview--kill-large-buffers ()
+  "Kill buffers previewing large files."
+  (mapc (lambda (pair)
+          (let ((buffer (cdr pair)))
+            (and (bufferp buffer)
+                 (kill-buffer buffer))))
+        dired-preview--large-files-alist)
+  (setq dired-preview--large-files-alist nil))
+
 (defun dired-preview--get-windows ()
   "Return windows that show previews."
   (seq-filter #'dired-preview--window-parameter-p (window-list)))
@@ -158,7 +180,10 @@ until it drops below this number.")
 (defun dired-preview--file-ignored-p (file)
   "Return non-nil if FILE extension is among the ignored extensions.
 See user option `dired-preview-ignored-extensions-regexp'."
-  (when-let ((ext (file-name-extension file)))
+  (when-let (((not (file-directory-p file)))
+             ((stringp dired-preview-ignored-extensions-regexp))
+             (ext (file-name-extension file :include-dot))
+             ((not (string-blank-p ext))))
     (string-match-p ext dired-preview-ignored-extensions-regexp)))
 
 (defun dired-preview--file-large-p (file)
@@ -167,8 +192,8 @@ See user option `dired-preview-ignored-extensions-regexp'."
 
 (defun dired-preview--file-displayed-p (file)
   "Return non-nil if FILE is already displayed in a window."
-  (when-let* ((buffer (get-file-buffer file))
-              (window (get-buffer-window buffer)))
+  (when-let ((buffer (get-file-buffer file))
+             (window (get-buffer-window buffer)))
     (window-live-p window)))
 
 (defun dired-preview--set-window-parameters (window value)
@@ -178,41 +203,138 @@ See user option `dired-preview-ignored-extensions-regexp'."
     (set-window-parameter window 'dedicated value)
     (set-window-parameter window 'no-other-window value)))
 
-(defun dired-preview--run-mode-hooks ()
-  "Run mode hooks in current buffer, if `delayed-mode-hooks' is non-nil."
-  (cond
-   ((window-parameter (selected-window) 'dired-preview-window)
-    (dired-preview--delete-windows))
-   ((and delay-mode-hooks (current-buffer))
+(defun dired-preview--clean-up-window ()
+  "Delete or clean up preview window."
+  (if (window-parameter (selected-window) 'dired-preview-window)
+      (dired-preview--delete-windows)
     (dired-preview--set-window-parameters (selected-window) nil)
-    (apply #'run-hooks (delete-dups delayed-mode-hooks))
-    (kill-local-variable 'delayed-mode-hooks)
-    (remove-hook 'post-command-hook #'dired-preview--run-mode-hooks :local))))
+    (remove-hook 'post-command-hook #'dired-preview--clean-up-window :local)))
 
-(defun dired-preview--find-file-no-select (file)
-  "Call `find-file-noselect' on FILE with appropriate settings."
-  ;; NOTE: I learnt about `non-essential' and `delay-mode-hooks' from
-  ;; Daniel Mendler's `consult' package, which imnplements a preview
-  ;; functionality as well (more sophisticated than mine):
-  ;; <https://github.com/minad/consult>.
-  (let ((inhibit-message t)
-        (enable-dir-local-variables nil)
-        (enable-local-variables :safe)
-        (non-essential t)
-        (delay-mode-hooks t))
-    (find-file-noselect file :nowarn)))
+;; TODO 2024-04-22: Add PDF type and concomitant method to display its buffer.
+(defun dired-preview--infer-type (file)
+  "Infer what type FILE is.
+Return a cons cell whose `car' is a symbol describing FILE and `cdr' is
+FILE."
+  (let ((ext (file-name-extension file :include-dot))
+        (file (expand-file-name file)))
+    (cond
+     ((and (not (string-empty-p ext))
+           (string-match-p ext dired-preview-ignored-extensions-regexp))
+      (cons 'ignore file))
+     ((dired-preview--file-large-p file)
+      (cons 'large file))
+     ((and (not (string-empty-p ext))
+           (string-match-p ext dired-preview-image-extensions-regexp))
+      (cons 'image file))
+     ((file-directory-p file)
+      (cons 'directory file))
+     (t
+      (cons 'text file)))))
+
+(cl-defgeneric dired-preview--get-buffer (file)
+  "Get a buffer for FILE.")
+
+;; FIXME 2024-04-22: We have a lot of repetitive code.  Can we expand
+;; a macro inside of a `cl-defmethod' or, alternatively, have a macro
+;; that returns the method with its implementation?
+(cl-defmethod dired-preview--get-buffer ((file (head text)))
+  "Get preview buffer for text FILE type."
+  (cl-letf (((symbol-function 'recentf-track-opened-file) #'ignore))
+    (let ((file (cdr file))
+          (inhibit-message t)
+          (enable-dir-local-variables nil)
+          (enable-local-variables :safe)
+          (non-essential t))
+      (find-file-noselect file :nowarn))))
+
+(defun dired-preview--add-truncation-message ()
+  "Add a message indicating that the previewed file is truncated."
+  (let ((end-ov (make-overlay (1- (point-max)) (point-max))))
+    (overlay-put
+     end-ov 'display
+     (propertize "\n--PREVIEW TRUNCATED--" 'face 'shadow))))
+
+(declare-function hexl-mode "hexl")
+(declare-function hexl-mode-exit "hexl" (&optional arg))
+
+(defun dired-preview-hexl-toggle ()
+  "Toggle preview between text and `hexl-mode'."
+  (interactive)
+  (dolist (win (dired-preview--get-windows))
+    (with-selected-window win
+      (if (eq major-mode 'hexl-mode)
+          (hexl-mode-exit)
+        (hexl-mode))
+      (dired-preview--add-truncation-message))))
+
+(cl-defmethod dired-preview--get-buffer ((file (head large)))
+  "Get preview buffer for large FILE.
+The size of the leading chunk is specified by
+`dired-preview-chunk-size'."
+  (cl-letf (((symbol-function 'recentf-track-opened-file) #'ignore))
+    (let ((file (cdr file))
+          (inhibit-message t)
+          (enable-dir-local-variables nil)
+          (enable-local-variables :safe)
+          (non-essential t))
+      (if-let* ((buffer (or (get-file-buffer file)
+                            (find-buffer-visiting file)
+                            (alist-get file dired-preview--large-files-alist
+                                       nil nil #'equal))))
+          buffer ; Buffer is already being visited, we can reuse it
+        (with-current-buffer (create-file-buffer file)
+          ;; We create a buffer with a partial preview
+          (buffer-disable-undo)
+          (insert-file-contents file nil 1 dired-preview-chunk-size 'replace)
+          (when (eq buffer-file-coding-system 'no-conversion)
+            (hexl-mode))
+          (dired-preview--add-truncation-message)
+          (read-only-mode t)
+          ;; Because this buffer is not marked as visiting FILE, we need to keep
+          ;; track of it ourselves.
+          (setf (alist-get file dired-preview--large-files-alist
+                           nil nil 'equal)
+                (current-buffer)))))))
+
+(cl-defmethod dired-preview--get-buffer ((file (head ignore)))
+  "Get preview buffer for ignored FILE."
+  (message "No preview method for `%s'" (cdr file)))
+
+(cl-defmethod dired-preview--get-buffer ((file (head directory)))
+  "Get preview buffer for directory FILE type."
+  (cl-letf (((symbol-function 'recentf-track-opened-file) #'ignore))
+    (let ((file (cdr file))
+          (inhibit-message t)
+          (enable-dir-local-variables nil)
+          (enable-local-variables :safe)
+          (non-essential t))
+      (dired-noselect file))))
+
+;; FIXME 2024-04-22: Best way to preview images and PDF files?  For now
+;; this is the same as the text file type, though we need to refine
+;; it.
+(cl-defmethod dired-preview--get-buffer ((file (head image)))
+  "Get preview buffer for image FILE type."
+  (cl-letf (((symbol-function 'recentf-track-opened-file) #'ignore))
+    (let ((file (cdr file))
+          (inhibit-message t)
+          (enable-dir-local-variables nil)
+          (enable-local-variables :safe)
+          (non-essential t))
+      (find-file-noselect file :nowarn))))
 
 (defun dired-preview--add-to-previews (file)
   "Add FILE to `dired-preview--buffers', if not already in a buffer.
 Always return FILE buffer."
-  (let ((buffer (find-buffer-visiting file)))
-    (if (buffer-live-p buffer)
-        buffer
-      (setq buffer (dired-preview--find-file-no-select file)))
-    (with-current-buffer buffer
-      (add-hook 'post-command-hook #'dired-preview--run-mode-hooks nil :local))
-    (add-to-list 'dired-preview--buffers buffer)
-    buffer))
+  (cl-letf (((symbol-function 'recentf-track-opened-file) #'ignore))
+    (let ((buffer (find-buffer-visiting file)))
+      (if (buffer-live-p buffer)
+          buffer
+        (setq buffer (dired-preview--get-buffer (dired-preview--infer-type file))))
+      (with-current-buffer buffer
+        (add-hook 'post-command-hook #'dired-preview--clean-up-window nil :local))
+      (add-to-list 'dired-preview--buffers buffer)
+      buffer)))
 
 (defun dired-preview--get-preview-buffer (file)
   "Return buffer to preview FILE in."
@@ -235,9 +357,10 @@ checked against `split-width-threshold' or
 
 (defun dired-preview-display-action-side ()
   "Pick a side window that is appropriate for the given frame."
-  (if-let* ((width (window-body-width))
-            ((>= width (window-body-height)))
-            ((>= width split-width-threshold)))
+  (if-let (split-width-threshold
+           (width (window-body-width))
+           ((>= width (window-body-height)))
+           ((>= width split-width-threshold)))
       `(:side right :dimension window-width :size ,(dired-preview-get-window-size :width))
     `(:side bottom :dimension window-height :size ,(dired-preview-get-window-size :height))))
 
@@ -251,7 +374,16 @@ aforementioned user option."
       (,(plist-get properties :dimension) . ,(plist-get properties :size)))))
 
 (defvar dired-preview-trigger-commands
-  '(dired-next-line dired-previous-line dired-mark dired-goto-file)
+  '( dired-next-line
+     dired-previous-line
+     dired-mark
+     dired-unmark
+     dired-unmark-backward
+     dired-del-marker
+     dired-goto-file
+     dired-find-file
+     scroll-up-command
+     scroll-down-command)
   "List of Dired commands that trigger a preview.")
 
 (defvar dired-preview--timer nil
@@ -266,13 +398,15 @@ aforementioned user option."
   "Kill preview buffers and delete their windows."
   (dired-preview--cancel-timer)
   (dired-preview--delete-windows)
-  (dired-preview--kill-buffers))
+  (dired-preview--kill-buffers)
+  (dired-preview--kill-large-buffers))
 
 (defun dired-preview--close-previews-outside-dired ()
-  "Call `dired-preview--close-previews' if BUFFER is not in Dired mode."
+  "Call `dired-preview--close-previews' if the current buffer is not in Dired mode."
   (unless (eq major-mode 'dired-mode)
     (dired-preview--close-previews)
-    (remove-hook 'window-state-change-hook #'dired-preview--close-previews-outside-dired)))
+    (remove-hook 'window-state-change-hook #'dired-preview--close-previews-outside-dired)
+    (put 'dired-preview-start 'function-executed nil)))
 
 (defun dired-preview--display-buffer (buffer)
   "Call `display-buffer' for BUFFER.
@@ -292,11 +426,17 @@ Only do it with the current major mode is Dired."
 
 (defun dired-preview--preview-p (file)
   "Return non-nil if FILE can be previewed."
-  (and (file-regular-p file)
-       (not (file-directory-p file))
+  (and file
+       (or (file-regular-p file) (file-directory-p file))
+       (file-readable-p file)
        (not (dired-preview--file-displayed-p file))
-       (not (dired-preview--file-ignored-p file))
-       (not (dired-preview--file-large-p file))))
+       (not (dired-preview--file-ignored-p file))))
+
+(defun dired-preview-start (file)
+  "Preview FILE instantly when invoking Dired."
+  (unless (get 'dired-preview-start 'function-executed)
+    (put 'dired-preview-start 'function-executed t)
+    (dired-preview-display-file file)))
 
 (defun dired-preview-trigger (&optional no-delay)
   "Trigger display of file at point after `dired-preview-trigger-commands'.
@@ -304,36 +444,46 @@ With optional NO-DELAY do not start a timer.  Otherwise produce
 the preview with `dired-preview-delay' of idleness."
   (add-hook 'window-state-change-hook #'dired-preview--close-previews-outside-dired)
   (dired-preview--cancel-timer)
-  (if-let* ((file (dired-file-name-at-point))
-            ((dired-preview--preview-p file))
-            ((memq this-command dired-preview-trigger-commands)))
+  (let* ((file (dired-file-name-at-point))
+         (preview (dired-preview--preview-p file)))
+    (cond
+     ((and preview (memq this-command dired-preview-trigger-commands))
       (if no-delay
           (dired-preview-display-file file)
         (setq dired-preview--timer
-              (run-with-idle-timer
-               dired-preview-delay
-               nil
-               #'dired-preview-display-file
-               file)))
+              (run-with-idle-timer dired-preview-delay nil #'dired-preview-display-file file))))
+     ((and file preview)
+      (dired-preview-start file))
+     ((and (not preview)
+           (memq this-command dired-preview-trigger-commands))
+      (dired-preview--delete-windows)))
     (dired-preview--close-previews-outside-dired)))
 
 (defun dired-preview-disable-preview ()
   "Disable Dired preview."
   (unless (eq major-mode 'dired-mode)
-    (error "Can only use `dired-preview' in Dired"))
+    (user-error "Can only use `dired-preview' in Dired"))
   (remove-hook 'post-command-hook #'dired-preview-trigger :local)
-  (dired-preview--close-previews))
+  (dired-preview--close-previews)
+  (put 'dired-preview-start 'function-executed nil))
 
 (defun dired-preview-enable-preview ()
   "Enable Dired preview."
   (unless (eq major-mode 'dired-mode)
-    (error "Can only use `dired-preview' in Dired"))
+    (user-error "Can only use `dired-preview' in Dired"))
   (add-hook 'post-command-hook #'dired-preview-trigger nil :local)
   (dired-preview-trigger :no-delay))
 
+(defvar dired-preview-mode-map
+  (let ((map (make-sparse-keymap)))
+    (define-key map (kbd "C-c C-c") #'dired-preview-hexl-toggle)
+    map)
+  "Key map for `dired-preview-mode'.")
+
 ;;;###autoload
 (define-minor-mode dired-preview-mode
   "Buffer-local mode to preview file at point in Dired."
+  :keymap dired-preview-mode-map
   :global nil
   (if dired-preview-mode
       (dired-preview-enable-preview)
